#include <glm/common.hpp>
#include <glm/ext/matrix_float4x4.hpp>
#include <glm/ext/matrix_transform.hpp>
#include <glm/trigonometric.hpp>
#include <cstring>
#include <iostream>
#include <fstream>
#include <glm/glm.hpp>
#include <cstdint>

#include <utils/utils.hpp>
#include <coil/segment.hpp>
#include <coil/coil.hpp>

// x, y, z, r, g, b
#define FIELDS_IN_POINT 6
#define DAYS_IN_YEAR (isLeap? 366.0f : 365.0f)

using glm::vec4;
using glm::vec3;
using glm::mat4;
using glm::radians;
using glm::rotate;
using glm::sin;
using glm::cos;
using glm::translate;

using std::endl;
using std::cout;
using std::cerr;

CoilSegment::CoilSegment(
        Coil* coil, 
        uint16_t year, 
        uint16_t height, 
        int shift, 
        uint32_t sliceDetalization, 
        bool isLeap
    ) :
    coil(coil), year(year), 
    height(height), shift(shift), 
    sliceDetalization(sliceDetalization), isLeap(isLeap) {
        init();
        calculate();
    }

void CoilSegment::init () {
    verticesLength = DAYS_IN_YEAR * FIELDS_IN_POINT * (sliceDetalization+2);
    vertices = (float *) calloc(verticesLength, sizeof(float));
    if (vertices == NULL) {
        cerr << "Allocation failed" << endl;
        exit(-1);
    }
    calculate();
    // cout << "Initialized CoilSegment" 
    //     << "(year: " << year 
    //     << ", height: " << height 
    //     << ", shift: " << shift
    //     << ", isLeap: " << isLeap <<")" << endl;
}

void CoilSegment::printVertices() {
    for (uint16_t day = 0; day < DAYS_IN_YEAR; day ++) {
        cout << "Day " << day << ", " <<  
            "x: " << vertices[FIELDS_IN_POINT * day + 0] << ", " <<
            "y: " << vertices[FIELDS_IN_POINT * day + 1] << ", " <<
            "z: " << vertices[FIELDS_IN_POINT * day + 2] << ", " <<
            "r: " << vertices[FIELDS_IN_POINT * day + 3] << ", " <<
            "g: " << vertices[FIELDS_IN_POINT * day + 4] << ", " <<
            "b: " << vertices[FIELDS_IN_POINT * day + 5] << endl;
    }
}


float *CoilSegment::calculateSlice() {
    uint32_t size = (/*start and end of a slize*/2 + sliceDetalization) * FIELDS_IN_POINT * sizeof(float);
    float *slice = (float *)malloc(size);
    mat4 transform = mat4(1.);
    transform = glm::scale(
        transform,
        vec3(
            coil->getWidth(),
            10,
            1
        )
    );
    vec3 point = vec3(cos(0), sin(0), 0);
    // Start of a slice
    slice[0] = cos(0);
    slice[1] = sin(0);
    slice[2] = 0;
    slice[3] = 1;
    slice[4] = 1;
    slice[5] = 1;

    float degreeByPoint = -180. / (sliceDetalization + 1);

    for (uint32_t i = 1; i <= sliceDetalization; i ++) {
        uint32_t offset = FIELDS_IN_POINT * i;
        
        slice[offset + 0] = round_to_presicion(cos(radians(degreeByPoint * i)), 5);
        slice[offset + 1] = round_to_presicion(sin(radians(degreeByPoint * i)), 5);
        slice[offset + 2] = 0;
        slice[offset + 3] = 1;
        slice[offset + 4] = 1;
        slice[offset + 5] = 1;
    }

    // End of a slice
    uint32_t endIndex = (size / sizeof(float) - FIELDS_IN_POINT); 
    slice[endIndex + 0] = round_to_presicion(cos(radians(180.)), 5);
    slice[endIndex + 1] = round_to_presicion(sin(radians(180.)), 5);
    slice[endIndex + 2] = 0;
    slice[endIndex + 3] = 1;
    slice[endIndex + 4] = 1;
    slice[endIndex + 5] = 1;

    return slice;
}

void CoilSegment::constructSegment(float *slice) {
    float degreesPerDay = 360.f / DAYS_IN_YEAR;

    for (uint32_t day = 0; day < DAYS_IN_YEAR; day ++) {
        float daysDegree = day * degreesPerDay;
        vec3 daysPosition = vec3(
            cos(radians(daysDegree)) * coil->getRadius(),
            10* daysDegree/360.,
            sin(radians(daysDegree)) * coil->getRadius()
        );

        mat4 transform = mat4(1.);

        transform = translate(
            transform,
            daysPosition
        );

        transform = rotate(
            transform,
            -radians(daysDegree),
            vec3(
                0.0,
                1.0,
                0.0
            )
        );

        for (uint32_t slicePoint = 0; slicePoint < (2 + sliceDetalization); slicePoint ++) {
            uint32_t slicePointOffset = FIELDS_IN_POINT * slicePoint;
            vec4 point(
                slice[slicePointOffset + 0],
                slice[slicePointOffset + 1],
                slice[slicePointOffset + 2],
                1
            );
            
            point = transform * point;
            uint64_t currentPointOffset = day * (2+sliceDetalization) * FIELDS_IN_POINT + slicePointOffset;
            vertices[currentPointOffset + 0] = point.x;
            vertices[currentPointOffset + 1] = point.y;
            vertices[currentPointOffset + 2] = point.z;

            vertices[currentPointOffset + 3] = 0.5;
            vertices[currentPointOffset + 4] = 0.5;
            vertices[currentPointOffset + 5] = 0.5;
            
            // memcpy(
            //     vertices + currentPointOffset + 3,
            //     slice + slicePointOffset + 3,
            //     3
            // );
        }
    }
}
void CoilSegment::fillVBO() {
    // printVertices();
    glBufferData(GL_ARRAY_BUFFER, sizeof(float) * verticesLength, vertices, GL_STATIC_DRAW);
}
uint64_t CoilSegment::fillEBO(uint64_t EBO) {
    uint64_t indicesLength = 6 * (sliceDetalization+2) * DAYS_IN_YEAR;
    uint64_t currentIndex = 0;
    uint64_t* indices = (uint64_t*) calloc(indicesLength, sizeof(uint64_t));
    for (uint64_t day = 0; day < DAYS_IN_YEAR; day ++) {
        for (uint64_t pointOfSlice = 0; pointOfSlice < (sliceDetalization+2); pointOfSlice ++) {
            // cout << currentIndex << endl;
            uint64_t dayOffset = day * (sliceDetalization+2);
            uint64_t offset = dayOffset + pointOfSlice;
            //First triangle
            indices[currentIndex] = offset;
            currentIndex ++;
            indices[currentIndex] = offset + (sliceDetalization+2);
            currentIndex ++;
            indices[currentIndex] = offset + 1;
            currentIndex++;
            //Second triangle
            
            if (pointOfSlice % (sliceDetalization+1) == 0 && pointOfSlice != 0) {
                indices[currentIndex] = offset - (sliceDetalization+2);
                currentIndex++;
                indices[currentIndex] = offset + 1;
                currentIndex++;
                indices[currentIndex] = offset;
            } else {
                indices[currentIndex] = offset + 1;
                currentIndex++;
                indices[currentIndex] = offset + (sliceDetalization+3);
                currentIndex ++;
                indices[currentIndex] = offset + (sliceDetalization+2);
            }
            currentIndex ++;            
        }
    }

    // for (uint64_t i = 0; i < indicesLength; i ++) {
    //     cout << indices[i] << ",";
    // }
    // cout << endl;
    // cout << indicesLength << endl;

    if (currentIndex != indicesLength) {
        cerr << "currentIndex == indicesLength assertion failed" << endl; 
        cerr << currentIndex << " " << indicesLength << endl;
    }
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, EBO);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(unsigned long) * indicesLength, indices, GL_STATIC_DRAW); 
    free(indices);
    return indicesLength;
}

// void testEBO(uint64_t *EBO, uint64_t length) {
//     for (uint64_t i = 0; i < length; i ++) {

//     }
// }
// Sample layout of 2 slices with 3 points.
//  |x |y |z |r |g |b |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |
//  0  1  2  3  4  5  6  7  8  9  10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35 36 37 38
//  |     point       |    point        |     point       |    point        |      point      |    point        |  
//  |                  slice                              |             slice                                   |
//  
void CoilSegment::calculate() {
    float *slice = calculateSlice();

    constructSegment(slice);
    exportSegmentToCsv();

}

float *CoilSegment::getVertices() { return vertices; }

uint64_t CoilSegment::getVerticesLength() { return verticesLength; }

uint16_t CoilSegment::getYear() { return year; }

void CoilSegment::render() {

}
void CoilSegment::exportSliceToCSV(float *slice) {
    cout << "Exporting slice to csv" << endl;
    std::ofstream outFile("./slice_export.csv");

    outFile << "x,y,z" << endl;
    
    for (uint32_t i = 0; i < sliceDetalization+2; i ++) {
        uint32_t offset = FIELDS_IN_POINT * i;
        outFile
                  << slice[offset + 0] << ","
                  << slice[offset + 1] << ",0" << endl;
    }
}

void CoilSegment::exportSegmentToCsv() {
    cout << "Exporting segment to csv" << endl;
    std::ofstream outFile("./segment_export.csv");

    outFile << "x,y,z" << endl;

    for (uint16_t day = 0; day < DAYS_IN_YEAR; day ++) {
        uint64_t dayOffset = day * (2+sliceDetalization) * FIELDS_IN_POINT;

        for (uint32_t point_in_slice = 0; point_in_slice < (2+sliceDetalization); point_in_slice ++) {
            uint32_t sliceOffset = point_in_slice * FIELDS_IN_POINT;
            outFile << vertices[dayOffset +  sliceOffset + 0] << "," 
                      << vertices[dayOffset +  sliceOffset + 1] << ","
                      << vertices[dayOffset +  sliceOffset + 2] << endl; 
        }
    }
}